import json
from urllib.parse import urljoin
from aiofreepybox.exceptions import *


class Access:
    def __init__(self, session, base_url, session_token, http_timeout):
        self.session = session
        self.header = {'X-Fbx-App-Auth': session_token}
        self.base_url = base_url
        self.timeout = http_timeout

    async def get(self, end_url):
        '''
        Send get request and return results
        '''
        url = urljoin(self.base_url, end_url)
        r = await self.session.get(url, headers=self.header, timeout=self.timeout)
        resp = await r.json()

        if not resp['success']:
            raise HttpRequestError('GET request failed (APIResponse: {0})'
                                   .format(json.dumps(resp)))

        return resp['result'] if 'result' in resp else None

    async def post(self, end_url, payload=None):
        '''
        Send post request and return results
        '''
        url = urljoin(self.base_url, end_url)
        data = json.dumps(payload) if payload is not None else None
        r = await self.session.post(url, headers=self.header, data=data, timeout=self.timeout)
        resp = await r.json()

        if not resp['success']:
            raise HttpRequestError('POST request failed (APIResponse: {0})'
                                   .format(json.dumps(resp)))

        return resp['result'] if 'result' in resp else None

    async def put(self, end_url, payload=None):
        '''
        Send post request and return results
        '''
        url = urljoin(self.base_url, end_url)
        data = json.dumps(payload) if payload is not None else None
        r = await self.session.put(url, headers=self.header, data=data, timeout=self.timeout)
        resp = await r.json()

        if not resp['success']:
            raise HttpRequestError('PUT request failed (APIResponse: {0})'
                                   .format(json.dumps(resp)))

        return resp['result'] if 'result' in resp else None



class Dhcp:

    def __init__(self, access):
        self._access = access

    async def get_config(self):
        '''
        Get DHCP configuration
        '''
        return await self._access.get('dhcp/config/')

    async def set_config(self, conf):
        '''
        Update a config with new conf dictionary
        '''
        self._access.put('dhcp/config/', conf)

    async def get_dynamic_dhcp_lease(self):
        '''
        Get the list of DHCP dynamic leases
        '''
        return await self._access.get('dhcp/dynamic_lease/')

    async def get_static_dhcp_lease(self):
        '''
        Get the list of DHCP static leases
        '''
        return await self._access.get('dhcp/static_lease/')



import asyncio
import hmac
import ipaddress
import os
import json
import logging
import socket
import ssl
import time
from urllib.parse import urljoin

import aiohttp

import aiofreepybox
from aiofreepybox.exceptions import *
from aiofreepybox.access import Access
from aiofreepybox.api.system import System
from aiofreepybox.api.dhcp import Dhcp
from aiofreepybox.api.switch import Switch
from aiofreepybox.api.lan import Lan
from aiofreepybox.api.wifi import Wifi
from aiofreepybox.api.fs import Fs
from aiofreepybox.api.call import Call
from aiofreepybox.api.connection import Connection


# Token file default location
token_filename = 'app_auth'
token_dir = os.path.dirname(os.path.abspath(__file__))
token_file = os.path.join(token_dir, token_filename)

# Default application descriptor
app_desc = {
    'app_id': 'aiofpbx',
    'app_name': 'aiofreepybox',
    'app_version': aiofreepybox.__version__,
    'device_name': socket.gethostname()
    }

logger = logging.getLogger(__name__)


class Freepybox:
    def __init__(self, app_desc=app_desc, token_file=token_file, api_version='v3', timeout=10):
        self.token_file = token_file
        self.api_version = api_version
        self.timeout = timeout
        self.app_desc = app_desc

    async def open(self, host, port):
        '''
        Open a session to the freebox, get a valid access module
        and instantiate freebox modules
        '''
        if not self._is_app_desc_valid(self.app_desc):
            raise InvalidTokenError('Invalid application descriptor')

        cert_path = os.path.join(os.path.dirname(__file__), 'freebox_certificates.pem')
        ssl_ctx = ssl.create_default_context()
        ssl_ctx.load_verify_locations(cafile=cert_path)

        conn = aiohttp.TCPConnector(ssl_context=ssl_ctx)
        self.session = aiohttp.ClientSession(connector=conn, raise_for_status=True)

        self._access = await self._get_freebox_access(host, port, self.api_version, self.token_file, self.app_desc, self.timeout)

        # Instantiate freebox modules
        self.system = System(self._access)
        self.dhcp = Dhcp(self._access)
        self.switch = Switch(self._access)
        self.lan = Lan(self._access)
        self.wifi = Wifi(self._access)
        self.fs = Fs(self._access)
        self.call = Call(self._access)
        self.connection = Connection(self._access)

    async def close(self):
        '''
        Close the freebox session
        '''
        if self._access is None:
            raise NotOpenError('Freebox is not open')

        await self._access.post('login/logout')
        await self.session.close()

    async def _get_freebox_access(self, host, port, api_version, token_file, app_desc, timeout=10):
        '''
        Returns an access object used for HTTP requests.
        '''

        base_url = self._get_base_url(host, port, api_version)

        # Read stored application token
        logger.info('Read application authorization file')
        app_token, track_id, file_app_desc = self._readfile_app_token(token_file)

        # If no valid token is stored then request a token to freebox api - Only for LAN connection
        if app_token is None or file_app_desc != app_desc:
            logger.info('No valid authorization file found')

            # Get application token from the freebox
            app_token, track_id = await self._get_app_token(base_url, app_desc, timeout)

            # Check the authorization status
            out_msg_flag = False
            status = None
            while(status != 'granted'):
                status = await self._get_authorization_status(base_url, track_id, timeout)

                # denied status = authorization failed
                if status == 'denied':
                    raise AuthorizationError('The app token is invalid or has been revoked')

                # Pending status : user must accept the app request on the freebox
                elif status == 'pending':
                    if not out_msg_flag:
                        out_msg_flag = True
                        print('Please confirm the authentification on the freebox')
                    await asyncio.sleep(1)

                # timeout = authorization failed
                elif status == 'timeout':
                    raise AuthorizationError('Authorization timed out')

            logger.info('Application authorization granted')

            # Store application token in file
            self._writefile_app_token(app_token, track_id, app_desc, token_file)
            logger.info('Application token file was generated : {0}'.format(token_file))


        # Get token for the current session
        session_token, session_permissions = await self._get_session_token(base_url, app_token, app_desc['app_id'], timeout)

        logger.info('Session opened')
        logger.info('Permissions: ' + str(session_permissions))

        # Create freebox http access module
        fbx_access = Access(self.session, base_url, session_token, timeout)

        return fbx_access

    async def _get_authorization_status(self, base_url, track_id, timeout):
        '''
        Get authorization status of the application token
        Returns:
            unknown 	the app_token is invalid or has been revoked
            pending 	the user has not confirmed the authorization request yet
            timeout 	the user did not confirmed the authorization within the given time
            granted 	the app_token is valid and can be used to open a session
            denied 	    the user denied the authorization request
        '''
        url = urljoin(base_url, 'login/authorize/{0}'.format(track_id))
        r = await self.session.get(url, timeout=timeout)
        resp = await r.json()
        return resp['result']['status']

    async def _get_app_token(self, base_url, app_desc, timeout=10):
        """
        Get the application token from the freebox
        Returns (app_token, track_id)
        """
        # Get authentification token
        url = urljoin(base_url, 'login/authorize/')
        data = json.dumps(app_desc)
        r = await self.session.post(url, data=data, timeout=timeout)
        resp = await r.json()

        # raise exception if resp.success != True
        if not resp.get('success'):
            raise AuthorizationError('Authorization failed (APIResponse: {0})'
                                     .format(json.dumps(resp)))

        app_token = resp['result']['app_token']
        track_id = resp['result']['track_id']

        return(app_token, track_id)

    def _writefile_app_token(self, app_token, track_id, app_desc, file):
        """
        Store the application token in g_app_auth_file file
        """
        d = {**app_desc, 'app_token': app_token, 'track_id': track_id}

        with open(file, 'w') as f:
            json.dump(d, f)

    def _readfile_app_token(self, file):
        """
        Read the application token in the authentication file.
        Returns (app_token, track_id, app_desc)
        """
        try:
            with open(file, 'r') as f:
                d = json.load(f)
                app_token = d['app_token']
                track_id = d['track_id']
                app_desc = {k: d[k] for k in ('app_id', 'app_name', 'app_version', 'device_name') if k in d}
                return (app_token, track_id, app_desc)

        except FileNotFoundError:
            return (None, None, None)

    async def _get_session_token(self, base_url, app_token, app_id, timeout=10):
        """
        Get session token from freebox.
        Returns (session_token, session_permissions)
        """
        # Get challenge from API
        challenge = await self._get_challenge(base_url, timeout)

        # Hash app_token with chalenge key to get the password
        h = hmac.new(app_token.encode(), challenge.encode(), 'sha1')
        password = h.hexdigest()

        url = urljoin(base_url, 'login/session/')
        data = json.dumps({'app_id': app_id, 'password': password})
        r = await self.session.post(url, data=data, timeout=timeout)
        resp = await r.json()

        # raise exception if resp.success != True
        if not resp.get('success'):
            raise AuthorizationError('Starting session failed (APIResponse: {0})'
                                     .format(json.dumps(resp)))

        session_token = resp.get('result').get('session_token')
        session_permissions = resp.get('result').get('permissions')

        return(session_token, session_permissions)

    async def _get_challenge(self, base_url, timeout=10):
        '''
        Return challenge from freebox API
        '''
        url = urljoin(base_url, 'login')
        r = await self.session.get(url, timeout=timeout)
        resp = await r.json()

        # raise exception if resp.success != True
        if not resp.get('success'):
            raise AuthorizationError('Getting challenge failed (APIResponse: {0})'
                                     .format(json.dumps(resp)))

        return resp['result']['challenge']

    def _get_base_url(self, host, port, freebox_api_version):
        '''
        Returns base url for HTTPS requests
        :return:
        '''
        return 'https://{0}:{1}/api/{2}/'.format(host, port, freebox_api_version)

    def _is_app_desc_valid(self, app_desc):
        '''
        Check validity of the application descriptor
        '''
        return all(k in app_desc for k in ('app_id', 'app_name', 'app_version', 'device_name'))



#-*- coding:utf-8 -*-

'''
Provides authentification and row access to Freebox using Freebox OS developer API.
Freebox API documentation : http://dev.freebox.fr/sdk/os/
'''

__version__ = '0.0.5'
__all__ = ['aiofreepybox']

from aiofreepybox.aiofreepybox import Freepybox



class Connection:

    def __init__(self, access):
        self._access = access

    async def get_status(self):
        '''
        Get connection status:
        '''
        return await self._access.get('connection/')

    async def get_config(self):
        '''
        Get connection configuration:
        '''
        return await self._access.get('connection/config/')

    async def set_config(self, conf):
        '''
        Update connection configuration:
        '''
        await self._access.put('connection/config/', conf)

    async def get_xdsl(self):
        '''
        Get xdsl infos:
        '''
        return await self._access.get('connection/xdsl/')

    async def get_ftth(self):
        '''
        Get ftth infos:
        '''
        return await self._access.get('connection/ftth/')



class Call:

    def __init__(self, access):
        self._access = access

    async def get_call_list(self):
        '''
        Returns the collection of all call entries
        '''
        return await self._access.get('call/log/')




class InvalidTokenError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class NotOpenError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class AuthorizationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class HttpRequestError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)



class Lan:

    def __init__(self, access):
        self._access = access

    async def get_config(self):
        '''
        Get Lan configuration
        '''
        return await self._access.get('lan/config/')

    async def set_config(self, conf):
        '''
        Update Lan config with conf dictionary
        '''
        await self._access.put('lan/config/', conf)

    async def get_interfaces(self):
        '''
        Get browsable Lan interfaces
        '''
        return await self._access.get('lan/browser/interfaces')

    async def get_hosts_list(self, interface='pub'):
        '''
        Get the list of hosts on a given interface
        '''
        return await self._access.get('lan/browser/{0}'.format(interface))

    async def get_host_information(self, host_id, interface='pub'):
        '''
        Get specific host informations on a given interface¶
        '''
        return await self._access.get('lan/browser/{0}/{1}'.format(interface, host_id))

    async def set_host_information(self, host_id, conf, interface='pub'):
        '''
        Update specific host informations on a given interface¶
        '''
        await self._access.put('lan/browser/{0}/{1}'.format(interface, host_id), conf)



class Switch:

    def __init__(self, access):
        self._access = access

    async def get_status(self):
        '''
        Get Switch status
        '''
        return await self._access.get('switch/status/')

    async def get_port_conf(self, port_id):
        '''
        Get port_id Port configuration
        '''
        return await self._access.get('switch/port/{0}'.format(port_id))

    async def set_port_conf(self, port_id, conf):
        '''
        Update port_id Port configuration with conf dictionary
        '''
        await self._access.put('switch/port/{0}'.format(port_id), conf)

    async def get_port_stats(self, port_id):
        '''
        Get port_id Port stats
        '''
        return await self._access.get('switch/port/{0}/{1}'.format(port_id, 'stats'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
This example can be run safely as it won't change anything in your box configuration
'''

import asyncio
from aiofreepybox import Freepybox


async def demo():
    # Instantiate Freepybox class using default application descriptor
    # and default token_file location
    fbx = Freepybox()

    # To find out the HTTPS host and port of your freebox, go to
    # http://mafreebox.freebox.fr/api_version

    # Connect to the freebox
    # Be ready to authorize the application on the Freebox if you use this
    # example for the first time
    await fbx.open(host='abcdefgh.fbxos.fr', port=1234)

    # Dump freebox configuration using system API
    # Extract temperature and mac address
    fbx_config = await fbx.system.get_config()
    print('Freebox temperature : {0}'.format(fbx_config['temp_sw']))
    print('Freebox mac address : {0}'.format(fbx_config['mac']))

    # Dump DHCP configuration using dhcp API
    fbx_dhcp_config = await fbx.dhcp.get_config()
    # Modify ip_range configuration
    fbx_dhcp_config['ip_range_start'] = '192.168.0.10'
    fbx_dhcp_config['ip_range_end'] = '192.168.0.50'
    # Send new configuration to the freebox. This line is commented to avoid any disaster.
    # await fbx.dhcp.set_config(fbx_dhcp_config)

    # Get the call list and print the last call entry
    fbx_call_list = await fbx.call.get_call_list()
    print(fbx_call_list[0])

    # Reboot your freebox. This line is commented to avoid any disaster.
    # await fbx.system.reboot()

    # Close the freebox session
    await fbx.close()

loop = asyncio.get_event_loop()
loop.run_until_complete(demo())
loop.close()



# -*- coding: utf-8 -*-

import codecs
import os
import re
from setuptools import setup, find_packages


# Method for retrieving the version is taken from the setup.py of pip itself:
# https://github.com/pypa/pip/blob/master/setup.py
here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get the long description from the README file
with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='aiofreepybox',
    version=find_version("aiofreepybox", "__init__.py"),
    packages=find_packages(),
    author='stilllman',
    author_email='luc_touraille@yahoo.fr',
    description='Provides asynchronous authentication and access to Freebox servers',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['aiohttp>=3,<4'],
    include_package_data=True,
    url='https://github.com/stilllman/freepybox/tree/aiofreepybox',
    keywords='freebox async',
    classifiers=[
        'Programming Language :: Python',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: Home Automation',
    ],
)



class Wifi:

    def __init__(self, access):
        self._access = access

    async def get_global_config(self):
        '''
        Get wifi global configuration:
        '''
        return await self._access.get('wifi/config/')


    async def set_global_config(self, conf):
        '''
        Update wifi global configuration:
        '''
        await self._access.put('wifi/config/', conf)

    async def get_ap_list(self):
        '''
        Get wifi access points list
        '''
        return await self._access.get('wifi/ap/')

    async def get_ap(self, ap_id):
        '''
        Get wifi access point with the specific id
        '''
        return await self._access.get('wifi/ap/{0}'.format(ap_id))

    async def set_ap(self, ap_id, conf):
        '''
        Update wifi access point with the specific id
        '''
        await self._access.get('wifi/ap/{0}'.format(ap_id), conf)

    async def get_ap_allowed_channel(self, ap_id):
        '''
        Get allowed channels of the wifi access point
        '''
        return await self._access.get('wifi/ap/{0}/allowed_channel_comb/'.format(ap_id))

    async def get_station_list(self, ap_id):
        '''
        Get the list of Wifi Stations associated to the AP
        '''
        return await self._access.get('wifi/ap/{0}/stations/'.format(ap_id))

    async def get_ap_neighbors(self, ap_id):
        '''
        Get the list of Wifi neighbors seen by the AP
        '''
        return await self._access.get('wifi/ap/{0}/neighbors/'.format(ap_id))



class System:

    def __init__(self, access):
        self._access = access

    async def get_config(self):
        '''
        Get system configuration:
        '''
        return await self._access.get('system/')

    async def reboot(self):
        '''
        Reboot freebox
        '''
        await self._access.post('system/reboot')



import base64
import os
import logging

logger = logging.getLogger(__name__)


class Fs:

    def __init__(self, access):
        self._access = access
        self._path = '/'

    def pwd(self):
        '''
        Returns the working directory
        '''
        return self._path

    async def cd(self, path):
        '''
        Changes the current directory
        '''
        if await self._path_exists(path):
            self._path = os.path.join(self._path, path)
        else:
            logger.error('{0} does not exist'.format(os.path.join(self._path, path)))

    async def _path_exists(self, path):
        '''
        Returns True if the path exists
        '''
        try:
            await self.get_file_info(os.path.join(self._path, path))
            return True
        except:
            return False

    async def ls(self):
        '''
        List directory
        '''
        return [i['name'] for i in await self.list_file(self._path)]

    async def get_tasks_list(self):
        '''
        Returns the collection of all tasks
        '''
        return await self._access.get('fs/tasks/')

    async def list_file(self, path):
        '''
        Returns the list of files for the given path
        '''
        path_b64 = base64.b64encode(path.encode('utf-8')).decode('utf-8')
        return await self._access.get('fs/ls/{0}'.format(path_b64))

    async def get_file_info(self, path):
        '''
        Returns information for the given path
        '''
        path_b64 = base64.b64encode(path.encode('utf-8')).decode('utf-8')
        return await self._access.get('fs/ls/{0}'.format(path_b64))
